from typing import List, runtime_checkable, Protocol
import typing as t
from blok.models import Option, NestedDict
from blok.service import Service
from dataclasses import dataclass
from blok.renderer import Renderer
import inspect

from blok.utils import check_service_compliance


T = t.TypeVar("T")


@dataclass
class InitContext:
    dependencies: t.Dict[str, "Blok"]
    kwargs: t.Dict[str, t.Any]

    def get_service(self, identifier: t.Type[T]) -> "T":
        try:
            if hasattr(identifier, "get_identifier"):
                return self.dependencies[identifier.get_identifier()]
            else:
                assert isinstance(
                    identifier, str
                ), "Identifier must be a string or a service/blok class"
                return self.dependencies[identifier]
        except KeyError:
            raise ValueError(
                f"Service with identifier {identifier} not found in dependencies"
            )


@dataclass
class ExecutionContext:
    docker_compose: NestedDict
    file_tree: NestedDict


@runtime_checkable
class Blok(Protocol):
    def get_blok_name(self) -> str: ...

    def get_identifier(self) -> str: ...

    def get_dependencies(self) -> List[str]: ...

    def entry(self, render: Renderer): ...

    def build(self, context: ExecutionContext): ...
    def preflight(self, init: InitContext): ...
    def get_options(self, name: str) -> List[Option]: ...


def inspect_dependable_params(function):
    signature = inspect.signature(function)
    parameters = signature.parameters
    dependencies = []
    dependency_mapper = {}

    for index, (name, param) in enumerate(parameters.items()):
        if index == 0:
            # This is the self parameter
            continue

        else:
            # Check if the parameter is **kwargs like

            cls = param.annotation

            if param.annotation == InitContext:
                dependency_mapper["__context__"] = name
                continue

            if hasattr(cls, "get_identifier"):
                dependencies.append(cls.get_identifier())
                dependency_mapper[cls.get_identifier()] = name
                continue

            dependency_mapper["__kwarg__" + str(index)] = name

    return dependencies, dependency_mapper


def service(service_identifier: t.Union[str, Service]):
    def decorator(cls):
        cls.__service_identifier__ = (
            service_identifier
            if isinstance(service_identifier, str)
            else service_identifier.get_identifier()
        )
        cls.__is_service__ = True

        if not hasattr(cls, "get_identifier"):
            cls.get_identifier = classmethod(lambda self: self.__service_identifier__)

        return cls

    return decorator


def build_mapped_preflight_function(preflight_function, dependency_mapper):
    def mapped_preflight_function(self, context):
        kwargs = {}
        for identifier, name in dependency_mapper.items():
            if identifier.startswith("__kwarg__"):
                try:
                    kwargs[name] = context.kwargs[name]
                except KeyError:
                    raise ValueError(
                        f"Missing required argument {name} in preflight function. Neither a dependency nor a context argument with the same name was found. Available arguments: {context.kwargs.keys()}"
                    )

            elif identifier == "__context__":
                kwargs[name] = context

            else:
                kwargs[name] = context.dependencies[identifier]

        try:
            preflight_function(self, **kwargs)
        except TypeError as e:
            raise TypeError(
                f"Error while running preflight function {preflight_function.__name__} with arguments {kwargs} {context.kwargs}"
            ) from e
        except Exception as e:
            raise e

    return mapped_preflight_function


def blok(
    service_identifier: t.Union[str, Service],
    blok_name: t.Optional[str] = None,
    options: t.Optional[List[Option]] = None,
    dependencies: t.Optional[List[str]] = None,
):
    def decorator(cls):
        try:
            cls.__service_identifier__ = (
                service_identifier
                if isinstance(service_identifier, str)
                else service_identifier.get_identifier()
            )
            cls.__is_blok__ = True
            cls.__dependencies__ = (
                dependencies or []
            )  # Cannot use set because we need to maintain order
            cls.__options__ = options or []

            if not hasattr(cls, "get_blok_name"):
                cls.get_blok_name = classmethod(
                    lambda cls: blok_name or cls.__name__.lower().replace("blok", "")
                )
            else:
                assert isinstance(
                    getattr(cls, "get_blok_name"), classmethod
                ), "get_blok_name must be a class method"

            if not hasattr(cls, "get_identifier"):
                cls.get_identifier = classmethod(lambda cls: cls.__service_identifier__)

            if not hasattr(cls, "preflight"):
                cls.preflight = lambda self, context: None
            else:
                init_function = getattr(cls, "preflight")
                init_dependencies, dependency_mapper = inspect_dependable_params(
                    init_function
                )
                cls.preflight = build_mapped_preflight_function(
                    init_function, dependency_mapper
                )
                for i in init_dependencies:
                    if i not in cls.__dependencies__:
                        cls.__dependencies__.append(i)

            if not hasattr(cls, "get_identifier"):
                cls.get_identifier = classmethod(lambda cls: cls.__service_identifier__)

            if not hasattr(cls, "get_dependencies"):
                cls.get_dependencies = lambda cls: cls.__dependencies__

            if not hasattr(cls, "get_options"):
                cls.get_options = lambda cls: cls.__options__

            if not hasattr(cls, "entry"):
                cls.entry = lambda self, renderer: None

            if not hasattr(cls, "build"):
                cls.build = lambda self, context: None

            if isinstance(service_identifier, str):
                pass
            elif inspect.isclass(service_identifier):
                check_service_compliance(cls, service_identifier)

            else:
                raise ValueError("Service must be a string or a service class")

        except Exception as e:
            raise ValueError(f"Error while creating blok {cls.__name__}") from e

        return cls

    return decorator



"""Blok is a tool for building and managing docker-compose projec

Blok utilized the dependency injection pattern to build and manage docker-compose projects.
Services can be built and managed using the blok CLI tool. In general
blok aims to autodiscover services and build them using the __blok__ magic method.



"""

from .blok import blok, InitContext, ExecutionContext, Option
from .service import service
from .renderer import Renderer, Panel
from .cli.create import create_cli


__all__ = [
    "blok",
    "service",
    "InitContext",
    "ExecutionContext",
    "Option",
    "Renderer",
    "Panel",
    "create_cli",
]



from .builder import build_cli
from blok.registry import BlokRegistry
from blok.cli.magic_registry import MagicRegistry
import rich_click as click
from blok.renderers.click import RichRenderer
from blok.renderer import Renderer
from rich import get_console
import typing as t


def make_cli(registry: BlokRegistry, renderer: Renderer):
    build = build_cli(registry, renderer)

    @click.command()
    def inspect():
        """Inspect the bloks available in the python environment"""

        click.echo("Available bloks:")
        for blok in registry.bloks.values():
            click.echo(blok.get_blok_name() + "\t\t\t" + blok.get_identifier())

    @click.group()
    @click.pass_context
    def cli(ctx):
        """Welcome to blok! A tool for building and managing docker-compose projects.

        Blok utilized your locally installed projects to build and manage docker-compose projects.
        Projects can register bloks into the blok registry using the __blok__ magic method.
        For more information, visit [link=https://arkitekt.live/bloks]https://arkitekt_next.live/bloks[/link]
        """

        pass

    cli.add_command(build, "build")
    cli.add_command(inspect, "inspect")

    return cli


def create_cli(*bloks, magic: bool = False, renderer: t.Optional[Renderer] = None):
    if not magic:
        reg = BlokRegistry(strict=True)
    else:
        reg = MagicRegistry("__blok__")

    for blok in bloks:
        reg.add_blok(blok)

    renderer = renderer or RichRenderer(get_console())

    return make_cli(reg, renderer)



from blok.registry import BlokRegistry
import rich_click as click
import typing as t
from pathlib import Path
import yaml
from functools import partial
import os
from blok.cli.entrypoint import entrypoint
from blok.renderer import Renderer


def configure(config_file_name: str, ctx, param, filename):
    """Configures the context with the default map"""
    config_path = Path(filename) / config_file_name

    try:
        with open(config_path, "r") as f:
            cfg = yaml.load(f, yaml.SafeLoader)
        ctx.default_map = cfg
    except FileNotFoundError:
        pass

    return filename


def reconfigure(ctx, param, flag):
    """Configures the context with the default map"""
    if flag:
        print("Forcing reconfiguration")
        ctx.default_map = {}
    return flag


def wrap_builder(
    builder_func: t.Callable,
    blok_registry: BlokRegistry,
    renderer: Renderer,
    config_file_name: str,
):
    def wrapped_builder(*args, **kwargs):
        "Builds a docker-compose project utilizing the bloks available in the python environment and locally" ""
        return builder_func(blok_registry, renderer, config_file_name, *args, **kwargs)

    func = wrapped_builder
    func = click.option(
        "--dry",
        "-dry",
        is_flag=True,
        default=False,
    )(func)
    func = click.option(
        "--use-bloks",
        "-b",
        type=click.Choice(blok_registry.bloks.keys()),
        multiple=True,
        default=list(blok_registry.bloks.keys()),
        help="Modules to use",
    )(func)
    func = click.option(
        "--discard_bloks",
        "-d",
        type=click.Choice(blok_registry.bloks.keys()),
        multiple=True,
        default=[],
        help="Modules to discard in the build",
    )(func)
    func = click.option(
        "--force",
        "-f",
        is_flag=True,
        default=False,
        callback=reconfigure,
    )(func)

    for option in blok_registry.get_click_options():
        func = option(func)

    func = click.argument(
        "path",
        type=click.Path(readable=True, writable=True),
        default=lambda: Path(os.getcwd()),
        callback=partial(configure, config_file_name),
    )(func)
    func = click.argument("blok", type=click.Choice(blok_registry.bloks.keys()))(func)
    func = click.option(
        "--force",
        "-f",
        is_flag=True,
        default=False,
        callback=reconfigure,
    )(func)
    func = click.option(
        "--yes",
        "-y",
        is_flag=True,
        default=False,
    )(func)

    func = click.command()(func)

    return func


def build_cli(
    blok_registry: BlokRegistry,
    renderer: Renderer,
    config_file_name: str = "__blok__.yml",
):
    return wrap_builder(entrypoint, blok_registry, renderer, config_file_name)



from typing import Dict, Any, Protocol
from blok import blok, InitContext, Option
from blok import service
from dataclasses import dataclass
import socket
from dataclasses import dataclass, field
from typing import Set
from blok.bloks.services.dns import DNSResult, DnsService



@dataclass
class DNSResult:
    hostnames: Set[str] = field(default_factory=set)
    ip_addresses: Set[str] = field(default_factory=set)

def get_interface_addresses(netifaces, interface, check_ip6=True):
    """
    Get IP addresses for a given network interface.
    """
    addresses = []
    try:
        interface_details = netifaces.ifaddresses(interface)
        if netifaces.AF_INET in interface_details:
            for addr_info in interface_details[netifaces.AF_INET]:
                addresses.append(addr_info['addr'])
        if check_ip6 and netifaces.AF_INET6 in interface_details:
            for addr_info in interface_details[netifaces.AF_INET6]:
                addresses.append(addr_info['addr'])
    except ValueError:
        # If the interface doesn't support AF_INET or AF_INET6 or any other issue
        pass
    return addresses

def perform_dns_lookup(ip_address):
    """
    Perform a DNS lookup to find the hostname for a given IP address.
    """
    try:
        hostname, _, ip_addresses = socket.gethostbyaddr(ip_address)
        return hostname, ip_addresses
    except (socket.herror, socket.gaierror):
        # If the DNS lookup fails, return None
        return None, None

def get_dns_result(netifaces, check_ip6=True):
    # Get the list of all interfaces
    interfaces = netifaces.interfaces()
    dns_result = DNSResult()
    hostnames = set()
    ip_addresses = set()
    
    for interface in interfaces:
        addresses = get_interface_addresses(netifaces, interface, check_ip6)
        for address in addresses:
            hostname, new_addresses = perform_dns_lookup(address)
            if hostname:
                hostnames.add(hostname)
            if new_addresses:
                for new_address in new_addresses:
                    ip_addresses.add(new_address)


    dns_result.hostnames = sorted(hostnames)
    dns_result.ip_addresses = sorted(ip_addresses)
    
    
    return dns_result



@blok(DnsService, options=[Option("check_ip_six", default=False)])
class NetifacesBlok:

    def get_dns_result(self):
        return self.result

    def preflight(self, check_ip_six):
        import netifaces
        self.result = get_dns_result(netifaces, check_ip_six)


    def build(self, cwd):
        pass





class BlokError(Exception):
    """Base class for blok errors."""


class DependencyNotFoundError(BlokError):
    """Raised when a dependency is not found."""


class TooManyBlokFoundError(BlokError):
    """Raised when too many bloks are found."""


class BlokInitializationError(BlokError):
    """Raised when a blok fails to initialize."""


class BlokBuildError(BlokError):
    """Raised when a blok fails to build."""


class ProtocolError(BlokError):
    """Base class for protocol errors."""


class ProtocolNotCompliantError(BlokError):
    """Raised when an instance does not comply with a protocol."""



from blok.diff import Diff, ChangedDiff, AddedKeyDiff, RemoveKeyDiff, ListChangedDiff
from typing import List
from blok.tree.models import YamlFile, Repo


def construct_diff_tree(diffs: List[Diff]):
    from rich.tree import Tree

    tree = Tree("Differences")

    for diff in diffs:
        if isinstance(diff, ChangedDiff):
            tree.add(
                f"[yellow]Will Modify:[/yellow] {diff.path}\n  [red]Old:[/red] {diff.old}\n  [green]New:[/green] {diff.new}"
            )
        elif isinstance(diff, AddedKeyDiff):
            tree.add(f"[green]Will Add:[/green] {diff.path}/{diff.new}")
        elif isinstance(diff, RemoveKeyDiff):
            tree.add(f"[red]Will Discard (not delete):[/red] {diff.path}/{diff.old}")
        elif isinstance(diff, ListChangedDiff):
            tree.add(
                f"[yellow]List Modified:[/yellow] {diff.path}\n  [red]Item:[/red] {diff.old}"
            )

    return tree


def add_nodes(tree, structure):
    if not isinstance(structure, dict):
        return
    for key, value in structure.items():
        if isinstance(value, YamlFile):
            subtree = tree.add(f"[blue]{key} - YAML[/blue]")
            add_nodes(subtree, value.values)

        elif isinstance(value, Repo):
            # Create directory and clone repo
            subtree = tree.add(f"[red]{key}: {value} - REPO[/red]")
            # TODO: Clone repo

        elif isinstance(value, dict):
            # Add folder node and recurse into it
            subtree = tree.add(key)
            add_nodes(subtree, value)
        elif isinstance(value, list):
            # Add folder node and recurse into it
            subtree = tree.add(key)
            for item in value:
                if isinstance(item, dict):
                    add_nodes(subtree, item)
                else:
                    subtree.add(item)
        else:
            tree.add(f"{key}: [green]{value}[/green]")


def construct_file_tree(tree_name: str, structure: dict):
    from rich.tree import Tree

    tree = Tree(tree_name)
    add_nodes(tree, structure)
    return tree
